/*
 *  tcp.c - functions logging TCP connections
 *
 *  Copyright (C) 1998-1999 Hugo Haas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <pwd.h>
#include <pthread.h>

#include "defines.h"
#include "tcp.h"
#include "netutils.h"
#include "log.h"
#include "filter.h"
#include "configuration.h"
#include "ident.h"

/* Socket */
int tcp_socket;

extern int log_details;

struct loginfo tcp_log;

/*
 * Structure of a TCP packet
 */

struct tcppacket {
  struct iphdr ip;
  struct tcphdr tcp;
};

/*
 * log_tcppacket & threaded_log_tcppacket
 *
 * Log a TCP connection
 */

void *log_tcppacket(void *pkt) {
  char remote_host[HOST_LENGTH];
  char service[SERVICE_LENGTH];

  *remote_host = '\0';

  /* Host filter */
  if (do_log(((struct tcppacket *) pkt)->ip.saddr,
             ((struct tcppacket *) pkt)->ip.daddr,
             remote_host,
             ((struct tcppacket *) pkt)->tcp.dest, IPPROTO_TCP) == TRUE) {
    char details[DETAILS_LENGTH];
    *details ='\0';
    /* Name resolution not done yet? */
    if (*remote_host == '\0')
      host_lookup(remote_host, ((struct tcppacket *) pkt)->ip.saddr);
    service_lookup("tcp", service, ((struct tcppacket *) pkt)->tcp.dest);
    if (log_details & RUN_TCP) {
      get_details(details,
                  ((struct tcppacket*) pkt)->ip.saddr,
                  ((struct tcppacket*) pkt)->tcp.source,
                  ((struct tcppacket*) pkt)->ip.daddr,
                  ((struct tcppacket*) pkt)->tcp.dest);
    }
    tcp_log.log(tcp_log.level_or_fd, "%s connection attempt from %s%s", service, remote_host, details);
  }
#ifdef _MULTITHREAD_
  free(pkt);
#endif
  return NULL;
}

#ifdef _MULTITHREAD_
void threaded_log_tcppacket(struct tcppacket pkt) {
  pthread_attr_t attr_t;
  pthread_t t;
  __u8 *image;

  image = (__u8 *) malloc(sizeof(struct tcppacket));
  memcpy(image, &pkt, sizeof(struct tcppacket));
  pthread_attr_init(&attr_t);
  pthread_attr_setdetachstate(&attr_t, PTHREAD_CREATE_DETACHED);
  pthread_create(&t, &attr_t, log_tcppacket, image);
  pthread_attr_destroy(&attr_t);
}
#endif

/*
 * log_tcp
 *
 * Main thread logging TCP connections
 *
 */

void *log_tcp(void *nobody) {
  struct tcppacket pkt;

  pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

  tcp_socket = socket(AF_INET, SOCK_RAW, IPPROTO_TCP);
  if (tcp_socket <= 0)
    exit(1);

  setuid(((struct passwd *)nobody)->pw_uid);

  for(;;) {
    if (read(tcp_socket, (struct tcppacket *) &pkt, sizeof(struct tcppacket)) == -1)
      exit(1);

    if (pkt.ip.ihl != 5) {
#ifdef _MULTITHREAD_
      threaded_log_suspicious((struct iphdr *) &pkt);
#else
      log_suspicious((void *) &pkt);
#endif
      continue;
    }

    if (pkt.tcp.syn == 1 && pkt.tcp.ack == 0) {
#ifdef _MULTITHREAD_
      threaded_log_tcppacket(pkt);
#else
      log_tcppacket((void *) &pkt);
#endif
    }
  }
  /*@NOTREACHED@*/ 
  return NULL;
}
